@@ -1,5 +1,6 @@
 /***************************************************************
  * Copyright (c) 2020 Xilinx, Inc.  All rights reserved.
+ * Copyright (C) 2023, Advanced Micro Devices, Inc. All Rights Reserved
  * SPDX-License-Identifier: MIT
  ***************************************************************/
 
@@ -39,6 +40,10 @@
 #define MAX_CMD_LEN		512U
 #define MAX_AES_KEY_LEN         64U
 #define PLATFORM_STR_LEN	128U
+#define FPGA_WORD_SIZE		4U
+#define FPGA_DUMMY_BYTE		0xFFU
+
+#define ZYNQMP_MAX_ERR	27U
 
 struct dfx_package_node {
 	int  flags;
@@ -65,6 +70,42 @@ typedef struct dfx_package_node FPGA_NODE;
 
 FPGA_NODE *head_node, *first_node, *temp_node = NULL, *prev_node, next_node;
 
+typedef struct {
+        int err_code;
+        char *err_str;
+} dfx_err;
+
+static dfx_err zynqmp_err[] = {
+	[0] = { .err_code =  XFPGA_ERROR_CSUDMA_INIT_FAIL, .err_str = "Failed to initialize the CSUDMA module" },
+	[1] = { .err_code =  XFPGA_ERROR_PL_POWER_UP, .err_str = " Failed to power-up the PL" },
+	[2] = { .err_code =  XFPGA_ERROR_PL_ISOLATION, .err_str = "Failed to perform the PS-PL isolation " },
+	[3] = { .err_code =  XPFGA_ERROR_PCAP_INIT, .err_str = "Failed to initialize the PCAP IP Initialization" },
+	[4] = { .err_code =  XFPGA_ERROR_BITSTREAM_LOAD_FAIL, .err_str = " PL Configuration failed" },
+	[5] = { .err_code =  XFPGA_ERROR_CRYPTO_FLAGS, .err_str = "Failed due to incorrect user crypto flags" },
+	[6] = { .err_code =  XFPGA_ERROR_HDR_AUTH, .err_str = "Failed to authenticate the image headers" },
+	[7] = { .err_code =  XFPGA_ENC_ISCOMPULSORY, .err_str = "Support only encrypted Image loading" },
+	[8] = { .err_code =  XFPGA_PARTITION_AUTH_FAILURE, .err_str = "Image authentication failed" },
+	[9] = { .err_code =  XFPGA_STRING_INVALID_ERROR, .err_str = "Firmware internal error" },
+	[10] = { .err_code =  XFPGA_ERROR_SECURE_CRYPTO_FLAGS, .err_str = "Failed due to incorrect user crypto flags" },
+	[11] = { .err_code =  XFPGA_ERROR_SECURE_MODE_EN, .err_str = "Supports only secure Images loading" },
+	[12] = { .err_code =  XFPGA_HDR_NOAUTH_PART_AUTH, .err_str = "Image header authentication failed" },
+	[13] = { .err_code =  XFPGA_DEC_WRONG_KEY_SOURCE, .err_str = "Decryption failed due to wrong key source" },
+	[14] = { .err_code =  XFPGA_ERROR_DDR_AUTH_VERIFY_SPK, .err_str = "DDR Image authentication failed due to Invalid keys" },
+	[15] = { .err_code =  XFPGA_ERROR_DDR_AUTH_PARTITION, .err_str = "Failed to authentication DDR image partition" },
+	[16] = { .err_code =  XFPGA_ERROR_DDR_AUTH_WRITE_PL, .err_str = "DDR Image authentication failed" },
+	[17] = { .err_code =  XFPGA_ERROR_OCM_AUTH_VERIFY_SPK, .err_str = "OCM Image authentication failed due to Invalid keys" },
+	[18] = { .err_code =  XFPGA_ERROR_OCM_AUTH_PARTITION, .err_str = "Failed to authentication OCM image partition" },
+	[19] = { .err_code =  XFPGA_ERROR_OCM_REAUTH_WRITE_PL, .err_str = "OCM Image authentication failed" },
+	[20] = { .err_code =  XFPGA_ERROR_PCAP_PL_DONE, .err_str = "Failed to get the PCAP done status" },
+	[21] = { .err_code =  XFPGA_ERROR_AES_DECRYPT_PL, .err_str = "Image AES decryption failed" },
+	[22] = { .err_code =  XFPGA_ERROR_CSU_PCAP_TRANSFER, .err_str = "PCAP failed to transfer the Image" },
+	[23] = { .err_code =  XFPGA_ERROR_PLSTATE_UNKNOWN, .err_str = "PL is in Unknow state" },
+	[23] = { .err_code =  XFPGA_ERROR_BITSTREAM_FORMAT, .err_str = "Bitstream format error" },
+	[24] = { .err_code =  XFPGA_ERROR_UNALIGN_ADDR, .err_str = "Error: Received Unaligned Bitstream address" },
+	[25] = { .err_code =  XFPGA_ERROR_AES_INIT, .err_str = "AES initialization failed" },
+	[26] = { .err_code =  XFPGA_ERROR_EFUSE_CHECK, .err_str = "Support only secure image configuration" },
+};
+
 static struct dfx_package_node *create_package(void);
 static struct dfx_package_node *get_package(int package_id);
 static int destroy_package(int package_id);
@@ -75,6 +116,11 @@ static int dfx_getplatform(void);
 static int find_key(struct dfx_package_node *package_node);
 static int lengthOfLastWord2(const char *input);
 static void strlwr(char *destination, const char *source);
+static int dfx_get_error(char *cmd);
+static void zynqmp_print_err_msg(int err);
+#ifdef ENABLE_LIBDFX_TIME
+static inline double gettime(struct timeval  t0, struct timeval t1);
+#endif
 
 /* Provide a generic interface to the user to specify the required parameters
  * for the library.The calling process must call this API before it performs
@@ -96,21 +142,28 @@ int dfx_cfg_init(const char *dfx_package_path,
 		  const char *devpath, unsigned long flags)
 {
 	FPGA_NODE *package_node;
+	int err, ret = 0;
 	int platform;
-	int ret, err;
 	size_t len;
+#ifdef ENABLE_LIBDFX_TIME
+	struct timeval t1, t0;
+	double time;
 
+	gettimeofday(&t0, NULL);
+#endif
 	platform = dfx_getplatform();
 	if (platform == INVALID_PLATFORM) {
 		printf("%s: fpga manager not enabled in the kernel Image\r\n",
 			__func__);
-		return -DFX_INVALID_PLATFORM_ERROR;
+		ret = -DFX_INVALID_PLATFORM_ERROR;
+		goto END;
 	}
 
 	package_node = create_package();
 	if (package_node == NULL) {
 		printf("%s: create_package failed\r\n", __func__);
-		return -DFX_CREATE_PACKAGE_ERROR;
+		ret = -DFX_CREATE_PACKAGE_ERROR;
+		goto END;
 	}
 
 	package_node->xilplatform = platform;
@@ -157,6 +210,12 @@ destroy_package:
 	err = destroy_package(package_node->package_id);
 	if (err)
 		printf("%s:Destroy package failed \r\n", __func__);
+END:
+#ifdef ENABLE_LIBDFX_TIME
+	gettimeofday(&t1, NULL);
+	time = gettime(t0, t1);
+	printf("%s API Time taken: %f Milli Seconds\n\r", __func__, time);
+#endif
 	return ret;
 }
 
@@ -172,20 +231,27 @@ destroy_package:
 int dfx_cfg_load(int package_id)
 {
 	FPGA_NODE *package_node;
-	int len, ret, fd, buffd;
+	int len, fd, buffd, ret = 0, err = 0;
 	char command[MAX_CMD_LEN];
 	char *str;
 	DIR *FD;
+#ifdef ENABLE_LIBDFX_TIME
+	struct timeval total_t1, total_t0, load_t1, load_t0;
+	double total_time, load_time;
 
+	gettimeofday(&total_t0, NULL);
+#endif
 	if (package_id < 0) {
 		printf("%s: Invalid package id\n", __func__);
-		return -DFX_INVALID_PACKAGE_ID_ERROR;
+		ret = -DFX_INVALID_PACKAGE_ID_ERROR;
+		goto END;
 	}
 
 	package_node = get_package(package_id);
 	if (package_node == NULL) {
 		printf("%s: fail to get package_node\n", __func__);
-		return -DFX_GET_PACKAGE_ERROR;
+		ret = -DFX_GET_PACKAGE_ERROR;
+		goto END;
 	}
 
 	if (!(package_node->flags & DFX_EXTERNAL_CONFIG_EN)) {
@@ -193,7 +259,8 @@ int dfx_cfg_load(int package_id)
 		if (fd < 0) {
 			printf("%s: Cannot open device file...\n",
 			       __func__);
-			return -DFX_FAIL_TO_OPEN_DEV_NODE;
+			ret = -DFX_FAIL_TO_OPEN_DEV_NODE;
+			goto END;
 		}
 
 		snprintf(command, sizeof(command),
@@ -228,21 +295,30 @@ int dfx_cfg_load(int package_id)
 	snprintf(command, sizeof(command), "echo -n %s > %s/path",
 		 package_node->load_image_dtbo_name,
 		 package_node->load_image_overlay_pck_path);
+#ifdef ENABLE_LIBDFX_TIME
+	gettimeofday(&load_t0, NULL);
+#endif
 	system(command);
+#ifdef ENABLE_LIBDFX_TIME
+	gettimeofday(&load_t1, NULL);
+#endif
 
 	if (!(package_node->flags & DFX_EXTERNAL_CONFIG_EN)) {
 		snprintf(command, sizeof(command),
 			 "cat /sys/class/fpga_manager/fpga0/state >> state.txt");
 		ret = dfx_state(command, "operating");
 		if (ret) {
+			err = dfx_get_error(command);
 			snprintf(command, sizeof(command), "rmdir %s",
 				 package_node->load_image_overlay_pck_path);
 			system(command);
-			printf("%s: Image configuration failed\n", __func__);
-			snprintf(command, sizeof(command),
-				 "cat /sys/class/fpga_manager/fpga0/state");
+			printf("%s: Image configuration failed with error: 0x%x\n",
+			        __func__, err);
+			if (package_node->xilplatform == ZYNQMP_PLATFORM)
+				zynqmp_print_err_msg(err);
 			system(command);
-			return -DFX_IMAGE_CONFIG_ERROR;
+			ret = -DFX_IMAGE_CONFIG_ERROR;
+			goto END;
 		}
 	}
 
@@ -254,10 +330,20 @@ int dfx_cfg_load(int package_id)
 			 package_node->load_image_overlay_pck_path);
 		system(command);
 		printf("%s: Image configuration failed\n", __func__);
-		return -DFX_IMAGE_CONFIG_ERROR;
+		ret = -DFX_IMAGE_CONFIG_ERROR;
+		goto END;
 	}
 
-	return 0;
+END:
+#ifdef ENABLE_LIBDFX_TIME
+	gettimeofday(&total_t1, NULL);
+	total_time = gettime(total_t0, total_t1);
+	load_time =  gettime(load_t0, load_t1);
+	printf("%s: Image load time from pre-allocated buffer: %f Milli Seconds\n\r",
+	       __func__, load_time);
+	printf("%s API Total time taken: %f Milli Seconds\n\r", __func__, total_time);
+#endif
+	return ret;
 }
 
 /* This API is Responsible for loading the drivers corresponding to a package
@@ -270,22 +356,31 @@ int dfx_cfg_drivers_load(int package_id)
 {
 	FPGA_NODE *package_node;
 	char command[MAX_CMD_LEN];
-	int len, ret;
+	int len, ret = 0;
 	char *str;
+#ifdef ENABLE_LIBDFX_TIME
+	struct timeval t1, t0;
+	double time;
 
+	gettimeofday(&t0, NULL);
+#endif
 	if (package_id < 0) {
 		printf("%s: Invalid package id\n", __func__);
-		return -DFX_INVALID_PACKAGE_ID_ERROR;
+		ret = -DFX_INVALID_PACKAGE_ID_ERROR;
+		goto END;
 	}
 
 	package_node = get_package(package_id);
 	if (package_node == NULL) {
 		printf("%s: fail to get package_node\n", __func__);
-		return -DFX_GET_PACKAGE_ERROR;
+		ret = -DFX_GET_PACKAGE_ERROR;
+		goto END;
 	}
 
-	if (package_node->load_drivers_dtbo_path == NULL)
-		return -DFX_NO_VALID_DRIVER_DTO_FILE;
+	if (package_node->load_drivers_dtbo_path == NULL) {
+		ret = -DFX_NO_VALID_DRIVER_DTO_FILE;
+		goto END;
+	}
 
 	snprintf(command, sizeof(command),
 		 "/configfs/device-tree/overlays/%s_driver_%d",
@@ -310,10 +405,16 @@ int dfx_cfg_drivers_load(int package_id)
 			 package_node->load_drivers_overlay_pck_path);
 		system(command);
 		printf("%s: Drivers DTBO config failed\n", __func__);
-		return -DFX_DRIVER_CONFIG_ERROR;
+		ret = -DFX_DRIVER_CONFIG_ERROR;
 	}
 
-	return 0;
+END:
+#ifdef ENABLE_LIBDFX_TIME
+	gettimeofday(&t1, NULL);
+	time = gettime(t0, t1);
+	printf("%s API Time taken: %f Milli Seconds\n\r", __func__, time);
+#endif
+	return ret;
 }
 
 /* This API is Responsible for unloading the drivers corresponding to a package
@@ -326,17 +427,25 @@ int dfx_cfg_remove(int package_id)
 {
 	FPGA_NODE *package_node;
 	char command[MAX_CMD_LEN];
+	int ret = 0;
 	DIR *FD;
+#ifdef ENABLE_LIBDFX_TIME
+	struct timeval t1, t0;
+	double time;
 
+	gettimeofday(&t0, NULL);
+#endif
 	if (package_id < 0) {
 		printf("%s: Invalid package id\n", __func__);
-		return -DFX_INVALID_PACKAGE_ID_ERROR;
+		ret = -DFX_INVALID_PACKAGE_ID_ERROR;
+		goto END;
 	}
 
 	package_node = get_package(package_id);
 	if (package_node == NULL) {
 		printf("%s: fail to get package_node\n", __func__);
-		return -DFX_GET_PACKAGE_ERROR;
+		ret = -DFX_GET_PACKAGE_ERROR;
+		goto END;
 	}
 
 	if (package_node->load_drivers_overlay_pck_path != NULL) {
@@ -360,7 +469,13 @@ int dfx_cfg_remove(int package_id)
 		}
 	}
 
-	return 0;
+END:
+#ifdef ENABLE_LIBDFX_TIME
+	gettimeofday(&t1, NULL);
+	time = gettime(t0, t1);
+	printf("%s API Time taken: %f Milli Seconds\n\r", __func__, time);
+#endif
+	return ret;
 }
 
 /* This API is Responsible for release/destroy the resouces allocated
@@ -374,16 +489,24 @@ int dfx_cfg_destroy(int package_id)
 {
 	FPGA_NODE *package_node;
 	char command[MAX_CMD_LEN];
+	int ret = 0;
+#ifdef ENABLE_LIBDFX_TIME
+	struct timeval t1, t0;
+	double time;
 
+	gettimeofday(&t0, NULL);
+#endif
 	if (package_id < 0) {
 		printf("%s: Invalid package id\n", __func__);
-		return -DFX_INVALID_PACKAGE_ID_ERROR;
+		ret = -DFX_INVALID_PACKAGE_ID_ERROR;
+		goto END;
 	}
 
 	package_node = get_package(package_id);
 	if (package_node == NULL) {
 		printf("%s: fail to get package_node\n", __func__);
-		return -DFX_GET_PACKAGE_ERROR;
+		ret = -DFX_GET_PACKAGE_ERROR;
+		goto END;
 	}
 
 	if (package_node->load_image_overlay_pck_path != NULL) {
@@ -401,7 +524,14 @@ int dfx_cfg_destroy(int package_id)
 		close_dma_buffer(package_node->dmabuf_info);
 	}
 
-	return destroy_package(package_node->package_id);
+	ret = destroy_package(package_node->package_id);
+END:
+#ifdef ENABLE_LIBDFX_TIME
+	gettimeofday(&t1, NULL);
+	time = gettime(t0, t1);
+	printf("%s API Time taken: %f Milli Seconds\n\r", __func__, time);
+#endif
+	return ret;
 }
 
 /* This API populates buffer with {Node ID, Unique ID, Parent Unique ID, Function ID}
@@ -415,17 +545,25 @@ int dfx_cfg_destroy(int package_id)
 int dfx_get_active_uid_list(int *buffer)
 {
 	const char* filename = "/sys/devices/platform/firmware:versal-firmware/uid-read";
-	int platform, count = 0;
+	int platform, ret = 0, count = 0;
 	FILE* fd;
+#ifdef ENABLE_LIBDFX_TIME
+	struct timeval t1, t0;
+	double time;
 
+	gettimeofday(&t0, NULL);
+#endif
 	platform = dfx_getplatform();
-	if (platform != VERSAL_PLATFORM)
-		return -DFX_INVALID_PLATFORM_ERROR;
+	if (platform != VERSAL_PLATFORM) {
+		ret = -DFX_INVALID_PLATFORM_ERROR;
+		goto END;
+	}
 
 	fd = fopen(filename, "rb");
 	if (!fd) {
 		printf("Unable to open file!");
-		return -DFX_FAIL_TO_OPEN_BIN_FILE;
+		ret = -DFX_FAIL_TO_OPEN_BIN_FILE;
+		goto END;
 	}
 
 	while(!feof(fd)) {
@@ -434,9 +572,14 @@ int dfx_get_active_uid_list(int *buffer)
 	}
 
 	fclose(fd);
-	count = (count - 1) *  sizeof(int);
-
-	return count;
+	ret = (count - 1) *  sizeof(int);
+END:
+#ifdef ENABLE_LIBDFX_TIME
+	gettimeofday(&t1, NULL);
+	time = gettime(t0, t1);
+	printf("%s API Time taken: %f Milli Seconds\n\r", __func__, time);
+#endif
+	return ret;
 }
 
 /* This API populates buffer with meta-header info related to the user
@@ -453,18 +596,26 @@ int dfx_get_meta_header(char *binfile, int *buffer, int buf_size)
 {
 	const char* filename = "/sys/devices/platform/firmware:versal-firmware/meta-header-read";
 	char command[2048], *token, *tmp, *tmp1;
-	int platform, count = 0;
+	int platform, ret = 0, count = 0;
 	FILE* fd;
 	DIR *FD;
+#ifdef ENABLE_LIBDFX_TIME
+	struct timeval t1, t0;
+	double time;
 
+	gettimeofday(&t0, NULL);
+#endif
 	platform = dfx_getplatform();
-	if (platform != VERSAL_PLATFORM)
-		return -DFX_INVALID_PLATFORM_ERROR;
+	if (platform != VERSAL_PLATFORM) {
+		ret = -DFX_INVALID_PLATFORM_ERROR;
+		goto END;
+	}
 
 	fd = fopen(binfile, "rb");
 	if (!fd) {
 		printf("Unable to open binary file!");
-		return -DFX_FAIL_TO_OPEN_BIN_FILE;
+		ret = -DFX_FAIL_TO_OPEN_BIN_FILE;
+		goto END;
 	}
 	fclose(fd);
 
@@ -488,7 +639,8 @@ int dfx_get_meta_header(char *binfile, int *buffer, int buf_size)
 	fd = fopen(filename, "rb");
 	if (!fd) {
 		printf("Unable to open sysfs binary file!");
-		return -DFX_FAIL_TO_OPEN_BIN_FILE;
+		ret = -DFX_FAIL_TO_OPEN_BIN_FILE;
+		goto END;
 	}
 
 	while(!feof(fd)) {
@@ -504,7 +656,14 @@ int dfx_get_meta_header(char *binfile, int *buffer, int buf_size)
 	fclose(fd);
 	closedir(FD);
 
-	return count * sizeof(int);
+	ret = count * sizeof(int);
+END:
+#ifdef ENABLE_LIBDFX_TIME
+	gettimeofday(&t1, NULL);
+	time = gettime(t0, t1);
+	printf("%s API Time taken: %f Milli Seconds\n\r", __func__, time);
+#endif
+	return ret;
 }
 
 static int read_package_folder(struct dfx_package_node *package_node)
@@ -793,12 +952,34 @@ static int dfx_state(char *cmd, char *state)
 	return 1;
 }
 
+static int dfx_get_error(char *cmd)
+{
+	char string[PLATFORM_STR_LEN];
+	FILE *fp;
+	int c;
+
+	system(cmd);
+	fp = fopen("state.txt", "r");
+	c = getc(fp);
+	while(c!=EOF) {
+		fscanf(fp, "%s", string);
+		c = getc(fp);
+	}
+
+	fclose(fp);
+
+	system("rm state.txt");
+
+    return (int)strtol(string, NULL, 0);
+}
+
 static int dfx_package_load_dmabuf(struct dfx_package_node *package_node)
 {
+	int word_align = 0, index, fd, ret;
 	struct dma_buf_sync sync = { 0 };
 	struct dma_buffer_info info;
 	long fileLen, count;
-	int fd, ret;
+	char *dma_buf;
 	FILE *fp;
 
 	fp = fopen(package_node->load_image_path, "rb");
@@ -812,16 +993,21 @@ static int dfx_package_load_dmabuf(struct dfx_package_node *package_node)
 	fileLen = ftell(fp);
 	fseek(fp, 0, SEEK_SET);
 
+	if (package_node->xilplatform == ZYNQMP_PLATFORM) {
+		word_align = fileLen % FPGA_WORD_SIZE;
+		if(word_align)
+			word_align = FPGA_WORD_SIZE - word_align;
+
+		fileLen = fileLen + word_align;
+	}
+
 	package_node->dmabuf_info = (struct dma_buffer_info *) calloc(1,
 					sizeof(struct dma_buffer_info));
 	package_node->dmabuf_info->dma_buflen = fileLen;
 
 	/* This call will do the following things
-	 *      1. Create an ION client
-	 *      2. Query ION heap_id_mask from ION heap
-	 *      3. Allocate memory for this ION client as per heap_type and
-	 *         return a valid buffer fd
-	 *      4. Create memory mapped buffer for the buffer fd
+	 *   1. Allocate memory from the DMA pool and return a valid buffer fd
+	 *   2. Create memory mapped buffer for the buffer fd
 	 */
 	ret = export_dma_buffer(package_node->dmabuf_info);
 	if (ret < 0) {
@@ -839,8 +1025,15 @@ static int dfx_package_load_dmabuf(struct dfx_package_node *package_node)
 	}
 
 	/* Copy Bitfile/PDI image into the Dmabuf */
-	count = fread((char *) package_node->dmabuf_info->dma_buffer, 1,
-		      fileLen, fp);
+	if (word_align) {
+		dma_buf = (char *) package_node->dmabuf_info->dma_buffer;
+		for (index = 0; index < word_align; index++)
+			dma_buf[index] = FPGA_DUMMY_BYTE;
+		fileLen = fileLen - word_align;
+		count = fread(&dma_buf[index], 1, fileLen, fp);
+	} else
+		count = fread((char *) package_node->dmabuf_info->dma_buffer, 1,
+			      fileLen, fp);
 	if (count != fileLen) {
 		printf("%s: Image copy failed\n", __func__);
 		goto unmap_buf;
@@ -944,3 +1137,39 @@ static inline void strlwr(char *destination, const char *source)
 		destination++;
 	}
 }
+
+static void zynqmp_print_err_msg(int err)
+{
+	int i, err_found = 0;
+
+	if ((err & 0xFF) == XFPGA_VALIDATE_ERROR)
+		printf("Error: Image validation");
+	if ((err & 0xFF) == XFPGA_PRE_CONFIG_ERROR)
+		printf("Error: Image Pre-configuration");
+	if ((err & 0xFF) == XFPGA_WRITE_BITSTREAM_ERROR)
+                printf("Error: Image write:");
+	if ((err & 0xFF) == XFPGA_POST_CONFIG_ERROR)
+                printf("Error: Image Post-configuration");
+	if ((err & 0xFF) == XFPGA_OPS_NOT_IMPLEMENTED)
+                printf("Error: Operation not supported");
+	if ((err & 0xFF) == XFPGA_INVALID_PARAM)
+                printf("Error: Invalid input parameters");
+
+	for (i = 0; i < ZYNQMP_MAX_ERR; i++) {
+		if (zynqmp_err[i].err_code == (err >> 8) & 0xFF) {
+			printf(": %s\r\n", zynqmp_err[i].err_str);
+			err_found++;
+			break;
+		}
+	}
+
+	if ((!err_found) && (err & 0xFF))
+		printf("\r\n");
+}
+
+#ifdef ENABLE_LIBDFX_TIME
+static inline double gettime(struct timeval  t0, struct timeval t1)
+{
+	return ((t1.tv_sec - t0.tv_sec) * 1000.0f + (t1.tv_usec -t0.tv_usec) / 1000.0f);
+}
+#endif
